use crossterm::{
	event::{read, Event, KeyCode},
	terminal, ExecutableCommand, cursor,
};
use std::panic;
use std::io::{stdout, Result, Write, Stdout};
use std::fs::read_to_string;


#[derive(Default)]
struct Nuudel {
	quitting: bool,
	cursorx: u16,
	cursory: u16,
	buffer: Buffer,
	filename: String,
}

#[derive(Default)]
struct Buffer {
	lines: Vec<String>,
}

fn main() -> Result<()> {
	// Setup panic handling
	let default_panic = panic::take_hook();
	panic::set_hook(Box::new(move |info| {
		let _ = Nuudel::restore_terminal();
		default_panic(info);
	}));

	// Basic stuff
	let mut nuudel = Nuudel::default();
	Nuudel::init_terminal()?;

	// Handle args
	let args: Vec<String> = std::env::args().collect();
	if let Some(arg) = args.get(1) {
		nuudel.open_file(arg)?;
	}

	nuudel.run(&mut stdout())?;
	Nuudel::restore_terminal()?;

	Ok(())
}

impl Nuudel {

	pub fn run(&mut self, stdout: &mut Stdout) -> Result<()> {
		self.draw_terminal(stdout)?;
		while !self.quitting {
			self.handle_events()?;
			self.draw_terminal(stdout)?;
		}
		Ok(())
	}

	fn init_terminal() -> Result<()> {
		terminal::enable_raw_mode()?;
		stdout().execute(terminal::EnterAlternateScreen)?;
		Ok(())
	}

	fn restore_terminal() -> Result<()> {
		stdout().execute(cursor::Show)?;
		stdout().execute(terminal::LeaveAlternateScreen)?;
		terminal::disable_raw_mode()?;
		Ok(())
	}
	fn quit(&mut self) {
		self.quitting = true;
	}

	fn handle_events(&mut self) -> Result<()> {
		match read()? {
			Event::Key(event) => match event.code {
				KeyCode::Esc => {
					self.quit();
				},
				KeyCode::Left => {
					if self.cursorx > 0 {
						self.cursorx -= 1;
						self.cursorx = self.cursorx.clamp(0, terminal::size()?.0-1);
					}
				},
				KeyCode::Right => {
					self.cursorx += 1;
					self.cursorx = self.cursorx.clamp(0, terminal::size()?.0-1);
				},
				KeyCode::Up => {
					if self.cursory > 0 {
						self.cursory -= 1;
					}
					self.cursory = self.cursory.clamp(0, terminal::size()?.1-1);
				},
				KeyCode::Down => {
					self.cursory += 1;
					self.cursory = self.cursory.clamp(0, terminal::size()?.1-1);
				},
				KeyCode::Char(_) => {
					self.buffer.lines[usize::try_from(self.cursory).unwrap()].insert(self.cursorx.into(), 'A');
					self.cursorx += 1;

				},
				_ => {}
			},
			_ => {}
		}
		Ok(())
	}

	fn draw_terminal(&mut self, stdout: &mut Stdout) -> Result<()> {
		//stdout.execute(terminal::BeginSynchronizedUpdate)?;
		stdout.execute(cursor::Hide)?;
		stdout.execute(cursor::MoveTo(0, 0))?;

		let mut y = 0;

		// Draw buffer one by one
		let mut lines = self.buffer.lines.iter();
		while y < terminal::size()?.1-1 {
			stdout.execute(terminal::Clear(terminal::ClearType::CurrentLine))?;
			if let Some(line) = lines.next() {
				write!(stdout, "{}", line)?;
			} else {
				write!(stdout, "~")?;
			}
			stdout.execute(cursor::MoveToNextLine(1))?;
			y += 1;
		}
		print!("Press ESC to quit");

		//stdout.execute(terminal::EndSynchronizedUpdate)?;
		stdout.execute(cursor::MoveTo(self.cursorx, self.cursory))?;
		stdout.execute(cursor::Show)?;

		Ok(())
	}

	fn open_file(&mut self, filename: &str) -> Result<()> {
		self.buffer.lines.clear();
		self.buffer.lines = read_to_string(filename)
			.unwrap()
			.lines()
			.map(String::from)
			.collect();
		self.filename = filename.to_string();
	Ok(())
	}
}
